import os, tkinter.filedialog, threading

from tkinter import *
from traceback import format_exc
from array import array
from time import time, sleep
from struct import unpack

curr_dir = os.path.abspath(os.curdir)

'''

Eternal Zunder: Zelda Classic 1.92 - 2.50 Quest Deprotector

    FOR PYTHON 3.0 and higher.  NOT 2.7 OR LOWER

    I made this because I wanted to fix some bugs in an abandoned
quest, but it was password protected so I couldn't even get to
it. I managed to write Eternal Zunder's decompression and inner
layer decryption without even bringing Z-Quest into IDA, but from
that I was only able to deprotect the temp files that Z-Quest
created while opening a quest.

    I wanted to allow others to do the same so they can learn from
others work, borrow sfx, gfx, scripts, and so abandoned quests
aren't left to rot. I say "Borrow" because we all know there is no
way someone is going to use someone else's assets and claim them
as their own, without you guys knowing who ACTUALLY made them.

    I never wanted to learn ASM, but I guess I got a crash course
in it. You guys (those who use this and aren't pissed it exists)
had better appreciate how much effort went into analyzing the
decryption routine's ASM and translating it into Python.



This was a project that took me about a week's time to complete

I started to experiment with deprotecting save files so they can
be edited just for shits and giggles, but I just didn't care
enough to finish it. The code is in place to decrypt them and
uncompress them, but not to recompress and re-encrypt them. If you
feed the program a .sav file it'll properly decrypt and decompress
it, but if you want to actually modify them and save them back,
then someone's gonna need to write a routine to re-compress and
re-encrypt them. 


    The older quests can't be deprotected for two reasons:
    
1: The algorithm the quests are compressed with is different than
the newer ones and I don't want to spend any more time on this
trying to figure it out. It looks like it's still LZSS, but the
flag byte's value makes no sense and I'm not going to try and
figure it out.

2: More importantly, the older quests have a different file
structure, and as such the password hash(or it might just be a
plaintext password) are not going to be in the same spot. Even if
I could replace it, I don't think that .qsu files existed at that
point. Because of that, if I were to deprotect the quest I'd have
to recompress and re-encrypt it; I'm not writing code for that.


Made by Moe

'''







"""######################################"""
#      LZSS DECOMPRESSION CONSTANTS
"""######################################"""

BUF_FILL_CHAR = 0

LEN_BITS = 4
LEN_MASK = (2**LEN_BITS)-1
OFF_BITS = 12
OFF_MASK = (2**OFF_BITS-1)<<LEN_BITS

HEADER_LEN = 4
THRESHOLD = 3

BUFFER_START = LEN_MASK + THRESHOLD

DICT_SIZE = 2**OFF_BITS

Range_8 = array("B", range(8))
Flag_Masks = array("B", [1, 2, 4, 8, 16, 32, 64, 128])


#it's not known which key a quest uses until you try it
"""IF NEW KEYS ARE ADDED TO ZC IN THE FUTURE, THEY CAN BE ADDED TO THIS
LIST ALSO, IF ANYONE KNOWS HOW THE COMPRESSION WORKS ON THE OLDER QUESTS,
FEEL FREE TO ADD IT. THE DECRYPTION KEYS ARE ALREADY HERE"""
Quest_Decryption_Keys = array("I", [1278576952,  #0x4C358938
                                    2444403409,  #0x91B2A2D1
                                    1249647495,  #0x4A7C1B87
                                    4181279206,  #0xF93941E6
                                    4245249684])#0xFD095E94

Quest_Decryption_Keys_A = array("H", [25321,  #0x62E9
                                      32020,  #0x7D14
                                       6786,  #0x1A82
                                        699,  #0x02BB
                                      57500])#0xE09C

Quest_Decryption_Keys_B = array("H", [13849,  #0x3619
                                      41579,  #0xA26B
                                      61500,  #0xF03C
                                      31506,  #0x7B12
                                      20111])#0x4E8F


#these are used to figure out if the quest is decrypting properly
Old_Quest_Match_String = bytearray(b'AG Zelda')
New_Quest_Match_String = bytearray(b'AG ZC En')
Save_File_Match_String = bytearray(b'slh!')

Save_File_Header_String = bytearray(b'Zelda Classic Save File')


class E_Zunder_Window(Tk):
    
    XOR_Key = array("B", [108, 111, 110, 103, 116,  97, 110])
    
    No_Pass_Hash = array("B", [212,  29, 140, 217,
                               143,   0, 178,   4,
                               233, 128,   9, 152,
                               236, 248,  66, 126])
    
    def __init__(self, **options):
        Tk.__init__(self, **options )
        
        self.title("Eternal Zunder: Zelda Classic 1.92+ Quest Deprotector")
        self.geometry("450x100+0+0")
        self.resizable(0, 0)
        
        self.menubar = Menu(self)
        self.menubar.add_command(label="Why on Earth would you do this?!?!?!", command=self._Show_About)
        self.config(menu=self.menubar)
        
        self.Input_Path = StringVar(self)
        self.Current_State = -1
        self.New_State = 0
        
        self.Processing = False
        self.Begin_Decryption = False
        
        self.Current_Job_Min = 0
        self.Current_Job_Max = 100
        self.Current_Progress = 0

        self.Idle_Update_Interval = 300
        

        #Create the input box
        self.Input_File_Field = Entry(self, textvariable=self.Input_Path)
        self.Input_File_Field.place(x=10, y=15, anchor=NW)
        self.Input_File_Field.config(width=59, state=DISABLED)


        #Add the buttons
        self.btn_browse = Button(self, text="Browse...", width=10, command=self._Make_Browse)
        self.btn_browse.place(x=375, y=12, anchor=NW)

        self.btn_browse = Button(self, text="Deprotect", width=10, command=self._Start_Deprotection)
        self.btn_browse.place(x=375, y=40, anchor=NW)

        self.Progress_Bar = Meter(self)
        self.Progress_Bar.place(x=10, y=39, anchor=NW)

        #Create and start the deprotect thread
        self.Deprotect_Thread = threading.Thread(target=self._Deprotect_Quest)
        self.Deprotect_Thread.daemon = True
        self.Deprotect_Thread.start()

        self.Displayed_Info_Text_Box = Text(self, height=2, bg='#ece9d8', state=NORMAL, width=64)
        self.Displayed_Info_Text_Box.place(x=-1, y=67, anchor=NW)

        #create the about window
        self.About_Window = E_Zunder_About(self)
        self.About_Window.Close()
        self._Idle_Update()


    def _Start_Deprotection(self):
        if not(self.Processing):
            self.New_State = 0
            self.Begin_Decryption = True


        
    #This function is to make the window to browse for the tags folder
    def _Make_Browse(self):
        temp = filedialog.askopenfilename(initialdir=curr_dir, title='Select the protected file')
        if len(temp):
            self.Input_File_Field.config(state=NORMAL)
            self.Input_File_Field.delete(0,END)
            self.Input_File_Field.insert(0,temp)
            self.Input_File_Field.config(state=DISABLED)


    def _Show_About(self):
        try:
            self.About_Window.update()
            self.About_Window.deiconify()
        except: pass



    def _Idle_Update(self):
        if self.Processing:
            self.Progress_Bar.Set_Value(( (self.Current_Progress - self.Current_Job_Min) /
                                          self.Current_Job_Max ))
        else:
            self.Progress_Bar.Set_Value(0.0, "")
            
        if self.Current_State == self.New_State:
            self.update_idletasks()
            self.after(self.Idle_Update_Interval, self._Idle_Update)
            return
        
        self.Current_State = self.New_State
        self.Displayed_Info_Text_Box.config(state=NORMAL)
        self.Displayed_Info_Text_Box.delete('0.0', END)


        if self.New_State == 0:
            self.Displayed_Info_Text_Box.insert(INSERT, 'This program will deprotect any Zelda Classic quest\n'+
                                                        'made in version 1.92 and higher, including v2.5.')
        elif self.New_State == 1:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Reading quest from file...')
            print('Reading quest from file...')
        elif self.New_State == 2:
            self.Displayed_Info_Text_Box.insert(INSERT, 'First decryption pass...')
            print('First decryption pass...')
        elif self.New_State == 3:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Second decryption pass...')
            print('Second decryption pass...')
        elif self.New_State == 4:
            self.Displayed_Info_Text_Box.insert(INSERT, "Decompressing...")
            print("Decompressing...")
        elif self.New_State == 5:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Finished!\nQuest password has been removed!')
            print('Finished!\nQuest password has been removed!')
        elif self.New_State == 6:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Finished!\nSave file has been decrypted and uncompresssed!')
            print('Finished!\nSave file has been decrypted and uncompresssed!')
        elif self.New_State == -1:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Could not locate the file specified.')
        elif self.New_State == -2:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Something went wrong with the password removal.\n'+
                                                        'Error report has been printed to the console.')
        elif self.New_State == -3:
            self.Displayed_Info_Text_Box.insert(INSERT, 'Could not locate matching keyset to decrypt quest.')

        elif self.New_State == -4:
            self.Displayed_Info_Text_Box.insert(INSERT, "Quest is too old. I don't know how the compression\n"+
                                                        "works on a quest this old, thus I cant unpack it. Sorry.")
            
        self.Displayed_Info_Text_Box.config(state=DISABLED)
        self.update_idletasks()
        self.after(self.Idle_Update_Interval, self._Idle_Update)


    def _Deprotect_Quest(self):
        """
           This function will load the file at the current input path,
           decrypt it, decompress it, remove the password, recompress it,
           re-encrypt it, write it to a new file.
        """
        while True:
            sleep(0.5)
            
            if not(self.Processing) and self.Begin_Decryption:
                self.Processing = True
                self.Begin_Decryption = False
                
                #get the input file path and make the slashes uniform
                Input_Path = self.Input_Path.get().replace('/', '\\')
                Output_Path = os.path.splitext((os.path.dirname(os.path.realpath(Input_Path)) + "\\DEPROTECTED\\" +
                                                os.path.basename(Input_Path) ))[0] + ".qsu"

                if os.path.isfile(Input_Path):
                    self.New_State = 1
                    
                    try:
                        '''OPEN THE ENCRYPTED FILE AND TURN IT INTO A MUTABLE ARRAY'''
                        with open(Input_Path, 'r+b') as in_file:
                            in_bytes = array("B", in_file.read())

                        key_set, new_format, is_save_file = self._Get_Key_Set_and_Is_New_Format(in_bytes)
                        
                        if key_set >= 0:
                            
                            if new_format:
                                self.New_State = 2
                                '''remove the outer layer of encryption'''
                                in_bytes = self._Main_Decrypt(in_bytes, key_set)

                                if is_save_file:
                                    Output_Path = os.path.splitext(Output_Path)[0] + ".sav"
                                else:
                                    self.New_State = 3
                                    '''remove the inner layer of encryption'''
                                    self._XOR_Decrypt(in_bytes)

                                
                                if not os.path.exists(os.path.dirname(Output_Path)):
                                    os.makedirs(os.path.dirname(Output_Path))

                            
                                '''OPEN THE OUTPUT FILE'''
                                with open(Output_Path, 'w+b') as out_file:
                                    self.New_State = 4
                                    '''decompress the file so it can be modified'''
                                    self._LZSS_Decompress(in_bytes, out_file)

                                    if not(is_save_file):
                                        """Replace the password hash with a "no-password" password hash"""
                                        self._Remove_Password(out_file)
                                        self.New_State = 5
                                    else:
                                        self.New_State = 6
                            else:
                                self.New_State = -4
                        else:
                            self.New_State = -3
                    except:
                        self.New_State = -2
                        print(format_exc())
                else:
                    self.New_State = -1
                    
                self.Processing = False


    def _Main_Decrypt(self, in_bytes, key_set):
        if in_bytes[:23] == Save_File_Header_String:
            Header_String_Len = 23
        else:
            Header_String_Len = 24
            
        #try to undo the outermost layer of encryption
        self.Current_Progress = 0
        self.Current_Job_Max = len(in_bytes)-(Header_String_Len+8)

        #array that will contain the decrypted bytes
        #we trim the header and footer hash off
        out_bytes = array("B", in_bytes[Header_String_Len+4:-4])


        #create the quest key and XOR it with the internal key
        Decrypt_Key = array("I", unpack(">I", in_bytes[Header_String_Len:Header_String_Len+4]) )
        Decrypt_Key[0] ^= Quest_Decryption_Keys[key_set]

        Key_Piece_A = Quest_Decryption_Keys_A[key_set]
        Key_Piece_B = Quest_Decryption_Keys_B[key_set]

        #arrays to hold temporary values
        """I am treating these as if they were the
        EAX, EBX, ECX, EDX, ESI, and EDI registers"""
        Reg_Vars = array("I", [0,0,0,0, 0,0])
        Temp_A_B = array("I", [0,0])

        #the number of sets of bytes to loop over and decrypt
        #this is multiplied by 2 so we dont have to in the loop
        Loop_Count = ((len(out_bytes)//2) + len(out_bytes)%2)*2

        try:
            #loop over the bytes in sets of 2 at a time
            for i in range(0, Loop_Count, 2):                    
                
                #if Temp_A will be negative
                if Decrypt_Key[0]&8388608:
                    Temp_A_B[0] = ((Decrypt_Key[0]>>8)&65535) + 4294901760
                else:
                    Temp_A_B[0] = (Decrypt_Key[0]>>8)&65535

                #if EDI will be negative
                if Temp_A_B[0]&2147483648:
                    Reg_Vars[5] = 4294967295
                else:
                    Reg_Vars[5] = 0

                #ECX
                Reg_Vars[2] = ( ((Decrypt_Key[0]<<9)|Reg_Vars[5])
                                + Decrypt_Key[0] + Key_Piece_A )&4294967295
                
                #Temp_B
                Temp_A_B[1] = Reg_Vars[2]&65535

                #if ECX will be negative
                if Decrypt_Key[0]&2147483648:
                    Reg_Vars[2] = 4294967295
                else:
                    Reg_Vars[2] = 0
                
                #EDI
                Reg_Vars[5] = (Reg_Vars[5] + Reg_Vars[2])&4294967295

                #if EDX will be negative
                if Decrypt_Key[0]&2147483648:
                    Reg_Vars[3] = (Decrypt_Key[0]>>16)+4294901760
                else:
                    Reg_Vars[3] = Decrypt_Key[0]>>16
                
                #EDX
                Reg_Vars[3] = (( ( Reg_Vars[3] + Reg_Vars[5] + Key_Piece_B)
                                 + (Reg_Vars[2] | (Temp_A_B[0]<<1)  )
                                ) & 65535)
                
                #if DX is negative
                if Reg_Vars[3]&32768:
                    Reg_Vars[3] = 4294901760 + Reg_Vars[3]


                #if Temp_B is negative
                if Temp_A_B[1]&32768:
                    Decrypt_Key[0] = ((Reg_Vars[3]<<16) + Temp_A_B[1] + 4294901760)&4294967295
                else:
                    Decrypt_Key[0] = ((Reg_Vars[3]<<16) + Temp_A_B[1])&4294967295
                
                #EDI
                Reg_Vars[5] = ((Temp_A_B[1]<<16) + Reg_Vars[3])&4294967295


                """decrypt and write the bytes"""
                out_bytes[i] = (out_bytes[i]^Reg_Vars[5])&255
                out_bytes[i+1] = ( (4294967296+out_bytes[i+1]) - Reg_Vars[5])&255

                
                self.Current_Progress = i
        except IndexError:
            #it's quicker to just catch the exception here
            pass
        except:
            print(format_exc())
        
        self.Current_Progress = len(in_bytes)-(Header_String_Len+8)
        return(out_bytes)



    def _Get_Key_Set_and_Is_New_Format(self, in_bytes):
        for key_set in (0, 1, 2, 3, 4):
            test_bytes = bytearray(self._Main_Decrypt(in_bytes[:50], key_set))
            
            if test_bytes[5:13] in (Old_Quest_Match_String, New_Quest_Match_String):
                if test_bytes[3] >= 116:
                    return(key_set, True, False)
                else:
                    return(key_set, False, False)
                
            elif test_bytes[0:4] == Save_File_Match_String:
                return(key_set, True, True)
            
            else:
                self._XOR_Decrypt(test_bytes)
                
                if test_bytes[5:13] in (Old_Quest_Match_String, New_Quest_Match_String):
                    if test_bytes[3] >= 116:
                        return(key_set, True, False)
                    else:
                        return(key_set, False, False)


        return(-1, False, False)



    def _XOR_Decrypt(self, in_bytes):
        #local variable for faster access
        XOR_Key = self.XOR_Key
        
        self.Current_Job_Max = len(in_bytes)
        self.Current_Progress = 0
        
        for i in range( len(in_bytes) ):
            #XOR the data with the key
            in_bytes[i] ^= XOR_Key[i%7]

            self.Current_Progress = i


    def _LZSS_Decompress(self, in_bytes, out_file):
        #current offset in the uncompressed data
        O_Pos = 0
        #current offset in the compressed data
        I_Pos = HEADER_LEN
        
        I_Size = len(in_bytes)

        #array to hold the dictionary bytes
        dict_bytes = array("B", [BUF_FILL_CHAR]*DICT_SIZE)

        #used to temporarely hold a decompressed byte before it's written
        write_buffer = array("B", [0])
        
        self.Current_Job_Max = I_Size
        self.Current_Progress = 0
        
        try:
                
            while I_Pos < I_Size:
                Flags = in_bytes[I_Pos]
                I_Pos += 1
                
                '''each bit in the flags determines how to copy the data'''
                for mask in Flag_Masks:
                    if Flags & mask:
                        '''if the flag is true then copy the byte directly from the input'''
                        dict_bytes[O_Pos%DICT_SIZE] = in_bytes[I_Pos]
                        write_buffer[0] = in_bytes[I_Pos]
                        out_file.write(write_buffer)
                        I_Pos += 1
                        O_Pos += 1
                    else:
                        '''if not then copy a stream of bytes from the dictionary buffer'''
                        Length = (in_bytes[I_Pos+1] & LEN_MASK) + THRESHOLD
                        Offset = ( in_bytes[I_Pos] + BUFFER_START +
                                  ((in_bytes[I_Pos+1] & OFF_MASK) << LEN_BITS) )

                        for j in range(Length):
                            dict_bytes[(O_Pos + j)%DICT_SIZE] = dict_bytes[(Offset + j)%DICT_SIZE]
                            write_buffer[0] = dict_bytes[(Offset + j)%DICT_SIZE]
                            out_file.write(write_buffer)

                        I_Pos += 2
                        O_Pos += Length
                
                self.Current_Progress = I_Pos
        except IndexError:
            #it's quicker to just catch the exception here
            pass
        except:
            print(format_exc())
        
        self.Current_Progress = self.Current_Job_Max


    def _Remove_Password(self, out_file):
        out_file.seek(46)
        out_file.write(self.No_Pass_Hash)


class E_Zunder_About(Toplevel):

    def __init__(self, Parent, **options):
        Toplevel.__init__(self, Parent, **options )
        
        self.Parent = Parent
        
        self.title("Call it what you will, but...")
        self.geometry("400x140")
        self.resizable(0, 0)
        self.protocol("WM_DELETE_WINDOW", self.Close)

        about_text = ("Friends, noobs, angry quest creators, lend me your eyes and ears.\n" +
                      "The storm has come, and it brings with it eternal lightning and thunder.\n\n" +
                      "This program was built on the ideal that knowledge should be shared freely.\n" +
                      "Remember, you are the one responsible for what you do with this, not me.\n\n" +
                      (" "*105) + "Sincerely,\n" +
                      (" "*105) + "~Moe")
        
        self.Text_Canvas = Canvas(self, width=400, height=140)
        self.Text_Canvas.pack()
        self.Text_Canvas.create_text((200, 70), text=about_text, width=380)

    def Close(self):
        self.withdraw()


class Meter(Frame):
    """
       This widget was taken from here and modified:
       https://mail.python.org/pipermail/python-list/2004-March/286176.html
    """
    def __init__(self, master, fillcolor='lightblue', text='', value=0.0, **kw):
        Frame.__init__(self, master, bg='#ece9d8', width=354, height=20)
        self.configure(**kw)
        
        self._c = Canvas(self, bg=self['bg'], highlightthickness=0,
                         width=self['width'], height=self['height'],
                         relief='sunken', bd=2)
        self._c.pack(fill='x', expand=1)
        
        self._r = self._c.create_rectangle(0, 0, 0, int(self['height'])+5,
                                           fill=fillcolor, width=0)
        self._t = self._c.create_text(int(self['width'])/2,
                                      int(self['height'])/2, text='')
        self.Set_Value(value, text)

    def Set_Value(self, value=0.0, text=None):
        if value < 0.0:
            value = 0.0
        elif value > 1.0:
            value = 1.0
        
        if text == None:
            #if no text is specified get the default percentage
            text = str(int(round(100 * value))) + ' %'
            
        self._c.coords(self._r, 0, 0,
                       (int(self['width'])+5) * value,
                       int(self['height'])+5)
        
        self._c.itemconfigure(self._t, text=text)


if __name__ == "__main__":
    Main_Window = E_Zunder_Window()
    Main_Window.mainloop()